import django_namespace_perms.util as nsp
from django_namespace_perms.constants import PERM_READ, PERM_WRITE

from django import template

register = template.Library()

@register.filter
def nsp_check_write(user, obj):
  return nsp.has_perms(user, obj, PERM_WRITE)

@register.filter
def nsp_check_read(user, obj):
  return nsp.has_perms(user, obj, PERM_READ)




from django.db import models
from django.conf import settings
from django.contrib.auth.models import Group

from django_namespace_perms.constants import *
from django_namespace_perms.util import NAMESPACES, obj_to_namespace
from django.db.models.signals import post_save, pre_delete

#############################################################################

def managed_perms_create(sender, **kwargs):
  inst = kwargs.get("instance")

  if not hasattr(inst, "nsp_managed"):
    return

  user_field_name, perms = inst.nsp_managed
  UserPermission.objects.create(
    user=getattr(inst, user_field_name),
    permissions=perms,
    namespace=obj_to_namespace(inst)
  )

def managed_perms_remove(sender, **kwargs):
  inst = kwargs.get("instance")
  if not hasattr(inst, "nsp_managed"):
    return

  user_field_name, perms = inst.nsp_managed

  perms = UserPermission.objects.filter(
    user=getattr(inst, user_field_name),
    namespace=obj_to_namespace(inst)
  ).all()

  for p in perms:
    p.delete()


def managed_perms(model):
  """
  Sets up a model to have automatically managed permissions. The only requirement
  is that the model has a foreignkey field linked to a user model and a 
  nsp_managed property that returns the field name of the user foreignkey
  as well as what perms to set in a tuple
  """

  post_save.connect(managed_perms_create, sender=model)
  pre_delete.connect(managed_perms_remove, sender=model)




#############################################################################

#class Group(models.Model):
#  name = models.CharField(max_length=255, unique=True)
#
#  class Meta:
#    db_table = u'nsp_group'
#
#  def __unicode__(self):
#    return self.name

#############################################################################

class GroupPermission(models.Model):
  group = models.ForeignKey(Group, blank=False)
  namespace = models.CharField(max_length=255, blank=False)
  permissions = models.IntegerField(choices=PERM_CHOICES, blank=False, default=PERM_READ)

  class Meta:
    db_table = u'nsp_group_permission'

  def __unicode__(self):
    return "%s: %s" % (self.group.name, self.namespace)

  def xbahn_replicate_relations(self):
    return {"group" : ("id", "name")}

#############################################################################

class UserPermission(models.Model):
  user = models.ForeignKey(settings.AUTH_USER_MODEL, blank=False)
  namespace = models.CharField(max_length=255, blank=False)
  permissions = models.IntegerField(choices=PERM_CHOICES, blank=False, default=PERM_READ)

  class Meta:
    db_table = u'nsp_user_permission'

  def __unicode__(self):
    return "%s: %s" % (self.user.username, self.namespace)

#############################################################################

class UserGroup(models.Model):
  user = models.ForeignKey(settings.AUTH_USER_MODEL, blank=False)
  group = models.ForeignKey(Group, blank=False)

  class Meta:
    db_table = u'nsp_user_group'
    unique_together = ('user', 'group')

  def xbahn_replicate_relations(self):
    return {"group" : ("id", "name")}

  def __unicode__(self):
    return "%s: %s" % (self.user.username, self.group.name)

#############################################################################

if hasattr(settings, 'XBAHN') and settings.XBAHN.get("replication") and "namespace_perms" in settings.XBAHN["replication"].get("replicate",[]):
    import twentyc.xbahn.django.replication as replication
    replication.replicate(
      UserPermission,
      GroupPermission,
      UserGroup
    )



from django.test import SimpleTestCase
from django.db import models

from django_namespace_perms import util, constants
import json

###############################################################################
"""
Set TEST_RUNNER to this class in settings for fast testing (no db creation)
"""

from django.test.runner import DiscoverRunner
class NoDbTestRunner(DiscoverRunner):
  def setup_databases(self, **kwargs):
    pass
  def teardown_databases(self, old_config):
    pass

###############################################################################

class ModelTestA(models.Model):
  @property
  def nsp_namespace(self):
    return "test.django_namespace_perms.modeltesta.1"

class ModelTestB(models.Model):
  allowedField = models.CharField(max_length=255)
  deniedField = models.CharField(max_length=255)

class ModelTestC(models.Model):
  @classmethod
  def nsp_namespace_base(cls):
    return "test.django_namespace_perms"

class ModelTestD(ModelTestC):
  @classmethod
  def nsp_namespace_create(cls):
    return "create.test.django_namespace_perms"

###############################################################################
# Create your tests here.

class NSPTestCase(SimpleTestCase):

  perms = {
    "django_namespace_perms.modeltestb.1" : constants.PERM_WRITE,
    "django_namespace_perms.modeltestb.2.allowedfield" : constants.PERM_READ,
    "a.b" : constants.PERM_READ,
    "a.b.c" : constants.PERM_READ | constants.PERM_WRITE,
    "a.b.d" : constants.PERM_DENY,
    "a.100" : constants.PERM_READ,
    "b" : constants.PERM_READ,
    "c.a.$" : constants.PERM_READ,
    "c.a.a" : constants.PERM_READ,

    "e.a" : constants.PERM_READ,
    "e.a.b" : constants.PERM_READ,

    "f.1" : constants.PERM_READ,
    "g.a.1" : constants.PERM_READ,
    "g.b" : constants.PERM_READ,
    "g.c" : constants.PERM_READ,
    "g.c.3" : constants.PERM_READ,

    "h" : constants.PERM_READ,

    "x" : constants.PERM_READ,
    "x.c" : constants.PERM_READ,
    "x.*.z" : constants.PERM_READ,
    "x.*.z.c" : constants.PERM_READ,
    "x.y.z.d" : constants.PERM_READ
  }

  def setUp(self):
    pass

  def test_namespace_override(self):
    obj = ModelTestA()
    obj.id = 1
    namespace = util.obj_to_namespace(obj)
    self.assertEqual(namespace, "test.django_namespace_perms.modeltesta.1")

  def test_namespace(self):
    obj = ModelTestB()
    obj.id=1
    namespace = util.obj_to_namespace(obj)
    self.assertEqual(namespace, "django_namespace_perms.modeltestb.1")

  def test_namespace_classmethods(self):
    self.assertEqual(util.obj_to_namespace(ModelTestC), "test.django_namespace_perms")
    self.assertEqual(util.obj_to_namespace(ModelTestD), "create.test.django_namespace_perms")

  def test_model_perms(self):
    obj = ModelTestB()
    obj.id = 1
    obj2 = ModelTestB()
    obj2.id = 2
    self.assertEqual(util.has_perms(self.perms, obj, constants.PERM_WRITE), True)
    self.assertEqual(util.has_perms(self.perms, obj2, constants.PERM_WRITE), False)
    self.assertEqual(util.has_perms(self.perms, [obj2, "allowedfield"], constants.PERM_READ), True)
    self.assertEqual(util.has_perms(self.perms, [obj2, "deniedfield"], constants.PERM_READ), False)
    self.assertEqual(util.has_perms(self.perms, [obj2, "allowedfield"], constants.PERM_WRITE), False)

  def test_permcode_to_namespace_view(self):
    label, flag = util.permcode_to_namespace("app.view_model")
    self.assertEqual("app.model.view", label)
    self.assertEqual(flag, constants.PERM_READ)

  def test_permcode_to_namespace_add(self):
    label, flag = util.permcode_to_namespace("app.add_model")
    self.assertEqual("app.model.add", label)
    self.assertEqual(flag, constants.PERM_WRITE)

  def test_permcode_to_namespace_delete(self):
    label, flag = util.permcode_to_namespace("app.delete_model")
    self.assertEqual("app.model.delete", label)
    self.assertEqual(flag, constants.PERM_WRITE)

  def test_permcode_to_namespace_change(self):
    label, flag = util.permcode_to_namespace("app.change_model")
    self.assertEqual("app.model.change", label)
    self.assertEqual(flag, constants.PERM_WRITE)

  def test_has_perms(self):
    self.assertEqual(util.has_perms(self.perms, "a.b", constants.PERM_READ), True)
    self.assertEqual(util.has_perms(self.perms, "a.b", constants.PERM_WRITE), False)
    self.assertEqual(util.has_perms(self.perms, "a.b.c", constants.PERM_WRITE), True)
    self.assertEqual(util.has_perms(self.perms, "a.b.c", constants.PERM_READ), True)
    self.assertEqual(util.has_perms(self.perms, "a.b.d", constants.PERM_READ), False)
    self.assertEqual(util.has_perms(self.perms, "a.b.d", constants.PERM_WRITE), False)
    
    self.assertEqual(util.has_perms(self.perms, "e.a.a", constants.PERM_READ, explicit=True), False)
    self.assertEqual(util.has_perms(self.perms, "e.a.b", constants.PERM_READ, explicit=True), True)


  def test_has_perms_wildcard(self):
    self.assertEqual(util.has_perms(self.perms, "b.*", constants.PERM_READ), True)
    self.assertEqual(util.has_perms(self.perms, "a.b.*", constants.PERM_READ, ambiguous=True), True)
    self.assertEqual(util.has_perms(self.perms, "d\..*", constants.PERM_READ, ambiguous=True), False)

  def test_perms_apply(self):
    data = {
      "a" : {
        "b": {
          "c" : "This should be here",
          "d" : "This should be gone"
        },
        "100" : "This should be here"
      },
      "b" : "This should be here",
      "c" : {
        "a" : {
          "b" : "This should be gone",
          "a" : "This should be here"
        }
      }
    }
    expected = {
      "a" : {
        "100" : data["a"]["100"],
        "b" : {
          "c" : data['a']['b']['c']
        }
      },
      "b" : data['b'],
      "c" : {
        "a" : {
          "a": data['c']['a']['a']
        }
      }
    }

    perms_struct = util.perms_structure(self.perms)
    result = util.permissions_apply(data, perms_struct)
    self.assertEqual(expected, result)

  def test_apply_with_permset(self):
    data = {
      "x" : {
        "a" : "This should be here",
        "b" : "This should be gone",
        "c" : "This should be here",
        "d" : "This should be here",
        "y": {
          "y" : {
            "a" : "This should be here",
            "b" : "This should be gone"
          },
          "z" : {
            "a" : "This should be here",
            "b" : "This should be gone",
            "c" : "This should be here",
            "d" : "This should be here"
          }
        }
      }
    }

    expected = {
      "x" : {
        "a" : data["x"]["a"],
        "c" : data["x"]["c"],
        "d" : data["x"]["d"],
        "y" : {
          "y" : {
            "a" : data["x"]["y"]["y"]["a"]
          },
          "z" : {
            "a" : data["x"]["y"]["z"]["a"],
            "c" : data["x"]["y"]["z"]["c"],
            "d" : data["x"]["y"]["z"]["d"]
          }
        }
      }
    }

    ruleset = {
      "require" : {
        "x.b" : 0x01,
        "x.c" : 0x01,
        "x.y.*.b" : 0x01,
        "x.y.z.c" : 0x01,
        "x.y.z.d" : 0x01
      }
    }

    perms_struct = util.perms_structure(self.perms)
    result = util.permissions_apply(data, perms_struct, debug=False, ruleset=ruleset)
    self.assertEqual(expected, result)

  def test_list_handlers(self):
    
    data = {
      "f" : [
        {"a" : 1, "b" : "should be here" },
        {"a" : 2, "b" : "should be gone" }
      ],
      "g" : {
        "a" : [
          {"a" : 1, "b" : "should be here" },
          {"a" : 2, "b" : "should be gone" }
        ],
        "b" : "should be here",
        "c" : [
          {"a" : 1, "b" : "should be gone" },
          {"a" : 2, "b" : "should be gone" },
          {"a" : 3, "b" : "should be here" }
        ]
      },
      "h" : {
        "a" : [
          {"a" : 1, "b" : "should be here" },
          {"a" : 2, "b" : "should be gone" }
        ],
        "b" : [
          {"a" : 1, "b" : "should be here" },
          {"a" : 2, "b" : "should be gone" }
        ]
      }
    }

    expected = {
      "f" : [
        data["f"][0]
      ],
      "g" : {
        "a" : [
          data["g"]["a"][0]
        ],
        "c" : [
          data["g"]["c"][2]
        ],
        "b" : data["g"]["b"]
      },
      "h" : {
        "a" : [data["h"]["a"][0]],
        "b" : [data["h"]["b"][0]]
      }
    }

    def namespace_builder(**kwargs):
      return str(kwargs.get("a"))

    def namespace_builder_absolute(**kwargs):
      return "g.a.%s" % kwargs.get("a")

    ruleset = {
      "require": {
        "g.c.1" : 0x01,
        "g.c.2" : 0x01,
        "g.c.3" : 0x01,
        "h.*.2" : 0x01
      },
      "list-handlers" : {
        "f" : {
          "namespace" : namespace_builder
        },
        "g.a" : {
          "namespace" : namespace_builder_absolute,
          "absolute" : True
        },
        "g.c" : {
          "namespace" : namespace_builder
        },
        "h.*" : {
          "namespace" : namespace_builder
        }
      }
    }

    perms_struct = util.perms_structure(self.perms)
    result = util.permissions_apply(data, perms_struct, debug=False, ruleset=ruleset)
    self.assertEqual(expected, result)
    

  def _test_performance(self):
    
    def mkdataset(depth=3):
      depth = depth - 1
      if depth <= 0:
        return
      return dict([(str(k),mkdataset(depth=depth)) for k in range(1,1000)])
    data = {
      "a" : mkdataset(3),
      "b" : mkdataset(3)
    }

    ruleset = {
      "require" : {
        "a.100.100" : 0x01,
      }
    }

    ruleset["require"].update(**dict([("b.100.%d" % i, 0x01) for i in range(1,100)]))

    import time
    
    perms_struct = util.perms_structure(self.perms)
    
    t= time.time()
    result = util.permissions_apply(data, perms_struct, debug=False)
    diff = time.time() - t
    print "\n\nPerformance test took: %.5f seconds" % (diff)
    
    t = time.time()
    result = util.permissions_apply(data, perms_struct, debug=False, ruleset=ruleset)
    diff = time.time() - t
    print "\n\nPerformance (w/ Explicit ruleset) test took: %.5f seconds" % (diff)




import re
import constants
import json
import inspect

from django.db.models.query import QuerySet


APP_NAMESPACES = [
]

NAMESPACES = [
]

#############################################################################

class PermissionFrame(object):
  def __init__(self, p):
#    print "PermissionFrame", p
    self.value = p

  def has_value(self):
    return (self.value is not None)

  def check(self, level):
    if self.value is None:
      return False
    return (self.value & level) != 0

#############################################################################

def dict_valid(v):
  return hasattr(v, "items") and callable(v.items)

def dict_from_namespace(keys, data):
  if type(keys) != list:
    keys = keys.split(".")
  root = p = d = {}
  j = 0
  for k in keys:
    d[k] = {}
    p = d
    d = d[k]
  p[k] = data
  return root

def dict_has_value(data, value):
  if data == value:
    return True
  for k, item in data.items():
    return dict_has_value(item, value)
  return False

def dict_get_path(data, keys):
  for k in keys:
    if data.has_key(k):
      data = data[k]
    else:
      return None
  return data


#############################################################################

def autodiscover_namespaces(*models):
  for model in models:
    ns = "%s.%s" % (model._meta.app_label, model._meta.model_name)
    NAMESPACES.append((ns,ns))
    if model._meta.app_label not in APP_NAMESPACES:
      APP_NAMESPACES.append(model._meta.app_label)
    for field in model._meta.fields:
      fld = field.name
      ns_ = "%s.*.%s" % (ns, fld)
      NAMESPACES.append((ns_, ns_))
  for app_ns in APP_NAMESPACES:
    NAMESPACES.append((app_ns, app_ns))
  NAMESPACES.sort()

#############################################################################

def load_perms(user):
  
  if hasattr(user, "_nsp_perms"):
    user._nsp_perms_struct = perms_structure(user._nsp_perms)
    return user._nsp_perms

  from django_namespace_perms.models import UserPermission, GroupPermission
  from django.conf import settings
  from django.core.exceptions import ObjectDoesNotExist
  from django.contrib.auth.models import Group

  if user.is_authenticated():
    perms = UserPermission.objects.filter(user=user)
    group_perms = GroupPermission.objects.filter(group__in=user.groups.all())
  else:
    # guest user
    if hasattr(settings, "NSP_GUEST_GROUP"):
      guest_group = settings.NSP_GUEST_GROUP
    else:
      guest_group = "Guest"
    try:
      group = Group.objects.get(name=guest_group)
      group_perms = GroupPermission.objects.filter(group=group)
    except ObjectDoesNotExist:
      group_perms = []
    perms = []

  permdict = {}
  for perm in group_perms:
    permdict[perm.namespace] = perm.permissions
  for perm in perms:
    permdict[perm.namespace] = perm.permissions

  user._nsp_perms = permdict
  user._nsp_perms_struct = perms_structure(permdict)

  return permdict

#############################################################################

def permcode_to_namespace(perm):
  label, perm_code = tuple(perm.split("."))
  a = re.match("(add|delete|change|view)_(.+)", perm_code)

  if a:
    if a.group(1) == "view":
      return ("%s.%s.%s" % (label, a.group(2), a.group(1)), constants.PERM_READ)
    else:
      return ("%s.%s.%s" % (label, a.group(2), a.group(1)), constants.PERM_WRITE)

  return (label, constants.PERM_READ)



#############################################################################

def obj_to_namespace(obj):
  namespace = str(obj)

  if inspect.isclass(obj):
    # class passed check existance of possible class methods
    if hasattr(obj, "nsp_namespace_create"):
      return obj.nsp_namespace_create().lower()
    elif hasattr(obj, "nsp_namespace_base"):
      return obj.nsp_namespace_base().lower()

  if hasattr(obj, "nsp_namespace"):
    return  obj.nsp_namespace.lower()


  if type(obj) == list:
    if len(obj) != 2:
      raise Exception("When passing a list to obj_to_namespace it is expected to have two items: The model instance and the field name")
    ns = "%s.%s" % (obj_to_namespace(obj[0]), obj[1].lower())
    return ns

  if hasattr(obj, "db_column") and hasattr(obj, "model"):
    namespace = "%s.%s.%s" % (
      obj.model._meta.app_label, 
      obj.model._meta.model_name,
      obj.name
    )
  elif hasattr(obj, "_meta"):
    namespace = "%s.%s" % (
      obj._meta.app_label, 
      obj._meta.model_name
    )
  if hasattr(obj, "id"):
    namespace = "%s.%s" % (namespace, obj.id)
 

  return namespace.lower()



#############################################################################

def has_perms(user, namespace, level, ambiguous=False, explicit=False):

  """
  Check if a user has perms to the specified name space.

  This should be the primary function you call to check if a user has access to
  something.

  user <dict|AUTH_USER_MODEL> - the user's permissions, can be a perm_structure dict
  a load_perms dict or the User model itself. Passing the user model may force a call
  of load_perms if they had not been loade yet.

  namespace <string|ModelInstance|list> - the namespace to check, can be the namespace
  string itself, the instance of a django model or a list holding the instance of a
  django model at index 0 and a model field name at index 1

  level <int> - permission level to check

  explicit <bool=False> - if true, explicit permissions are required to the
  full path provided in namespace, partial namespace matches will be ignored.
  """
  
  if type(namespace) not in [str, unicode]:
    if hasattr(namespace, "nsp_require_explicit"):
      explicit = namespace.nsp_require_explicit
    namespace = obj_to_namespace(namespace)

  if type(user) != dict:
    if user.is_superuser:
      return True
    perms = load_perms(user)
    permstruct = user._nsp_perms_struct
  else:
    perms = user
    if not perms.has_key("__ps"):
      permstruct = perms_structure(perms)
    else:
      permstruct = perms

  if explicit:
    return get_perms(permstruct, namespace.split("."), explicit=True).check(level)
  return get_perms(permstruct, namespace.split(".")).check(level)


#############################################################################

def check_perms(perms, prefix, ambiguous=False):

  """
  DEPRECATED - replaced by get_perms

  Return the user's perms for the specified prefix

  perms <dict> permissions dict
  prefix <string> namespace to check for perms

  ambiguous <bool=False> if True reverse wildcard matching is active and a perm check for a.b.* will
  be matched by the user having perms to a.b.c or a.b.d - only use this if you know what 
  you are doing.
  """

  try:

    token = prefix.split(".")

    i = 1
    l = len(token)
    r = 0

    # collect permission rules with a wildcard in them, so we dont do unecessary
    # regex searches later on
    perms_wc = {}
    for ns, p in perms.items():
      if ns.find("*") > -1:
        perms_wc[re.escape(ns).replace("\*", "[^\.]+")] = p

    while i <= l:
      k = ".".join(token[:i])
      matched = False

      # check for exact match
      if perms.has_key(k):
        r = perms.get(k)

      # check for wildcard matches (if any wildcard rules exist)
      elif perms_wc:
        for ns, p in perms_wc.items():
          a = "^%s$" % ns
          b = "^%s\." % ns
          j = len(a)
          u = len(b)
          if j > matched and re.match(a, k):
            r = p
            matched = j
          elif u > matched and re.match(b, k):
            r = p
            matched = u

      # if not matched at all and ambiguous flag is true, do ambiguous matching
      if not matched and ambiguous:
        m = "^%s" % re.escape(k).replace("\*", "[^\.]+")
        for ns, p in perms.items():
          if re.match(m, ns) and p > r:
            r = p
            break

      i += 1

    return r
  except:
    raise  

###############################################################################

def permissions_apply_additive(data, perms_struct):
  if not perms_struct:
    return {}
  
  if not dict_valid(perms_struct) or not dict_valid(data):
    return data

  rv = {}
  for k,v in data.items():
    if perms_struct.get(k):
      rv[k] = permissions_apply_additive(
        v, 
        perms_struct.get(k)
      )
    if perms_struct.get("@%s"%k):
      d = permissions_apply_additive(
        v, 
        perms_struct.get("@%s"%k)
      )
      if dict_valid(d) and dict_valid(rv.get(k)):
        rv[k].update(d)
      else:
        rv[k] = d
    if perms_struct.get("*"):
      d = permissions_apply_additive(
        v, 
        perms_struct.get("*")
      )
      if dict_valid(d) and dict_valid(rv.get(k)):
        rv[k].update(d)
      else:
        rv[k] = d

  return rv

#############################################################################

def permissions_apply_subtractive(data, perms_struct, debug=False):
  if not dict_valid(data):
    raise Exception("Wanted a dict or dict-like object got %s" % type(data))

  for k,p in perms_struct.items():
    if k[0] == "@":
      k = k[1:]

    if data.has_key(k):

      if not p or (dict_valid(data[k]) and not any(data[k])):
        del data[k]
      elif type(data[k]) == dict and type(p) == dict:
        permissions_apply_subtractive(
          data.get(k), 
          p, 
          debug=debug
        )
        
    elif k == "*":
      for n,v in data.items():
        if not p or (dict_valid(data[n]) and not any(data[n])):
          del data[n]
        elif dict_valid(p) and dict_valid(v):
          permissions_apply_subtractive(
            v, 
            p, 
            debug=debug
          )

#############################################################################

def get_perms(d, keys, explicit=False):
  t = type(d)

  #print "checking for",keys,"in",d,explicit

  if t == int or t == long:
    if explicit and keys:
      return PermissionFrame(None)
    return PermissionFrame(d)
  if not keys:
    return PermissionFrame(None)

  r_a = 0 
  r_b = 0
  k = keys[0]
  if d.has_key(k):
    r_a = get_perms(d[k], keys[1:], explicit=explicit)
    if r_a.has_value():
      return r_a

  if d.has_key("@%s" % k):
    r_c = get_perms(d["@%s" % k], keys[1:], explicit=explicit)
    if r_c.has_value():
      return r_c
  
  if d.has_key("*"):
    r_b = get_perms(d["*"], keys[1:], explicit=explicit)
    if r_b.has_value():
      return r_b

  return PermissionFrame(None)
 

def perms_struct_has_explicit_rule(d, keys, level):
  return get_perms(d, keys, explicit=True).check(level)
 
#############################################################################

def permissions_apply_ruleset_require_explicit(data, path, perm, perms_struct, full_path=None):
  d = data
  j = p = None
  a = 0
  
  if type(path) is not list:
    keys = path.split(".")
  else:
    keys = path

  for k in keys:
    if k == "*":
      if not dict_valid(d):
        return
      for i,j in d.items():
        permissions_apply_ruleset_require_explicit(
          j, keys[a+1:], perm, perms_struct, full_path=keys[:a]+[i]
        )
      return
    else:
      if dict_valid(d) and d.has_key(k):
        p = (d,k)
        d = d[k]
      else:
        return

    a += 1
  if full_path is None:
    r = perms_struct_has_explicit_rule(perms_struct, keys, perm)
  else:
    r = perms_struct_has_explicit_rule(perms_struct, full_path + keys, perm)
  if not r and p:
    del p[0][p[1]]
    

#############################################################################

def permissions_apply_list_handler(data, path, handler, perms_struct, ruleset={}, full_path=None, container=None):

  d = data
  a = 0
  
  if type(path) is not list:
    keys = path.split(".")
  else:
    keys = path

  l = len(keys)

  for k in keys:
    if k == "*":
      if dict_valid(d): 
        for i,j in d.items():
          
          if a < l-1:
            _keys = keys[a+1:]
          else:
            _keys = []
          permissions_apply_list_handler(
            j, _keys, handler, perms_struct, full_path=keys[:a]+[i],container=(d,i), ruleset=ruleset
          )
        return
      elif type(d) is not list:
        return
    else:
      if dict_valid(d) and d.has_key(k):
        container = (d,k)
        d = d[k]
      else:
        return

    a += 1

  if d:
    n = []
    rs = {}
    if ruleset.has_key("require"):
      rs.update(require=ruleset.get("require"))
    rs.update(handler.get("ruleset",{}))
    if full_path:
      _path = full_path + keys
    else:
      _path = keys

    for item in d:
     
      if dict_valid(item):
        final_path = handler.get("namespace")(**item).lower().split(".")
      else:
        final_path = handler.get("namespace")(obj=item).lower().split(".")

      if not handler.get("absolute"):
        final_path = _path + final_path

      #print final_path, item
      
      r = permissions_apply(
        dict_from_namespace(final_path, item), 
        perms_struct, 
        ruleset=rs
      )
      r = dict_get_path(r, final_path)
      if r:
        n.append(r)

    container[0][container[1]] = n



#############################################################################

def permissions_apply(data, perms_struct, path='', debug=False, ruleset=None):
  if not dict_valid(perms_struct):
    load_perms(perms_struct)
    perms_struct = perms_struct._nsp_perms_struct

  rv = permissions_apply_additive(data, perms_struct)

  if debug:
    print json.dumps(rv, indent=2)

  permissions_apply_subtractive(rv, perms_struct)

  if debug:
    print json.dumps(rv, indent=2)

  # apply ruleset
  if ruleset:
    for key, perm in ruleset.get("require", {}).items():
      permissions_apply_ruleset_require_explicit(rv, key, perm, perms_struct)
    for key, hdl in ruleset.get("list-handlers", {}).items():
      permissions_apply_list_handler(rv, key, hdl, perms_struct, ruleset=ruleset)
    
  if debug:
    print json.dumps(rv, indent=2)

  return rv

#############################################################################

def permissions_apply_to_serialized_model(smodel, perms_struct, data=None, ruleset={}):
  if hasattr(smodel, "instance"):
    inst = smodel.instance
  else:
    inst = smodel

  namespace_str = obj_to_namespace(inst)
  namespace = namespace_str.split(".")
  structure = d = {}

  if hasattr(inst, "nsp_ruleset"):
    ruleset.update(inst.nsp_ruleset)

  l = len(namespace)
  i = 0
  while i < l:
    k = namespace[i]
    if i < l-1:
      d[k] = {}
      d = d[k]
    elif data is None:
      d[str(k)] = smodel.data
    else:
      d[str(k)] = data
    i += 1
  
  # prepare ruleset
  if ruleset:
    _ruleset = {}
    for section, rules in ruleset.items():
      _ruleset[section] = {}
      for rule, perms in rules.items():
        _ruleset[section]["%s.%s" % (namespace_str, rule)] = perms
    ruleset = _ruleset

  r = permissions_apply(
    structure,
    perms_struct,
    ruleset=ruleset
  )

  for k in namespace:
    r = r.get(k,{})

  return r

#############################################################################

def perms_structure(perms):
  perms_wc = {"__ps":True}
  for ns, p in perms.items():
    pieces = ns.split(".")
    a = prev = perms_wc 
    n = 0 
    l = len(pieces)
    for k in pieces:
      if n < l-1:
        if not a.has_key(k):
          a[k] = {}
        elif type(a[k]) != dict:
          a["@%s"%k] = a[k]
          a[k] = {}
        a = a[k]
      else:
        if not a.has_key(k):
          a[k] = p
        else:
          a["@%s"%k] = p
      n += 1
   

  return perms_wc 



from rest_framework import permissions, serializers

from django_namespace_perms.util import (
  has_perms, 
  obj_to_namespace,
  permissions_apply_to_serialized_model
)

from django_namespace_perms.constants import PERM_READ, PERM_WRITE 
import logging
from exceptions import PermissionDenied


log = logging.getLogger(__name__)

class BasePermission(permissions.BasePermission):

  def debug(self, msg):
    print msg
    log.debug(msg)
  
  def has_permission(self, request, view):
    """
    self.debug("Check permission: %s, %s" % (request.method, view.model))
    if request.method == "POST":
      return has_perms(request.user, view.model, PERM_WRITE)
    else:
      return True
    """
    # since instance of object does not exist yet and we have no access to the
    # serializer data we need to handle POST permission checks during
    # PermissionedModelSerializer.create - always return true here.
    return True

  def has_object_permission(self, request, view, obj):
    self.debug("Check Object permissions %s, %s, %s" % (
      request.method,
      request.user,
      obj_to_namespace(obj)
    ))

    if request.method in permissions.SAFE_METHODS:
      return has_perms(request.user, obj, PERM_READ)
    else:
      return has_perms(request.user, obj, PERM_WRITE)

class PermissionedModelSerializer(serializers.ModelSerializer):

  def create(self, validated_data):
    if hasattr(self, "nsp_namespace_create"):
      user = self.context.get("request").user
      if user and has_perms(user, self.nsp_namespace_create(validated_data), PERM_WRITE):
        return serializers.ModelSerializer.create(self, validated_data)
      else:
        raise PermissionDenied("User not set in serializer context, or user does not have write permissions")
    else:
      raise PermissionDenied("Serializer missing classmethod '%s' - so we have no way to determine permissioning namespace for instance creation" % "nsp_namespace_create")
      

  def to_representation(self, instance):
    """
    Apply permissions to serialized data before sending it out for
    good
    """
    r = super(serializers.ModelSerializer, self).to_representation(instance)
    req = self.context.get("request",None)
    user = self.context.get("user")

    if not user and req:
      user = req.user

    if user:
      
      # superusers can see everything
      if user.is_superuser:
        return r

      r = permissions_apply_to_serialized_model(
        instance,
        user,
        data=r
      )
    return r



import re
import constants
import json


APP_NAMESPACES = [
]

NAMESPACES = [
]

#############################################################################

def autodiscover_namespaces(*models):
  for model in models:
    ns = "%s.%s" % (model._meta.app_label, model._meta.model_name)
    NAMESPACES.append((ns,ns))
    if model._meta.app_label not in APP_NAMESPACES:
      APP_NAMESPACES.append(model._meta.app_label)
    for field in model._meta.fields:
      fld = field.name
      ns_ = "%s.*.%s" % (ns, fld)
      NAMESPACES.append((ns_, ns_))
  for app_ns in APP_NAMESPACES:
    NAMESPACES.append((app_ns, app_ns))
  NAMESPACES.sort()

#############################################################################

def load_perms(user):
  
  if hasattr(user, "_nsp_perms"):
    user._nsp_perms_struct = perms_structure(user._nsp_perms)
    return user._nsp_perms

  from django_namespace_perms.models import UserPermission, GroupPermission
  from django.conf import settings
  from django.core.exceptions import ObjectDoesNotExist
  from django.contrib.auth.models import Group

  if user.is_authenticated():
    perms = UserPermission.objects.filter(user=user)
    group_perms = GroupPermission.objects.filter(group__in=user.groups.all())
  else:
    # guest user
    if hasattr(settings, "NSP_GUEST_GROUP"):
      guest_group = settings.NSP_GUEST_GROUP
    else:
      guest_group = "Guest"
    try:
      group = Group.objects.get(name=guest_group)
      group_perms = GroupPermission.objects.filter(group=group)
    except ObjectDoesNotExist:
      group_perms = []
    perms = []

  permdict = {}
  for perm in group_perms:
    permdict[perm.namespace] = perm.permissions
  for perm in perms:
    permdict[perm.namespace] = perm.permissions

  print permdict

  user._nsp_perms = permdict
  user._nsp_perms_struct = perms_structure(permdict)

  return permdict

#############################################################################

def permcode_to_namespace(perm):
  label, perm_code = tuple(perm.split("."))
  a = re.match("(add|delete|change|view)_(.+)", perm_code)

  if a:
    if a.group(1) == "view":
      return ("%s.%s.%s" % (label, a.group(2), a.group(1)), constants.PERM_READ)
    else:
      return ("%s.%s.%s" % (label, a.group(2), a.group(1)), constants.PERM_WRITE)

  return (label, constants.PERM_READ)



#############################################################################

def obj_to_namespace(obj):
  namespace = str(obj)
  if hasattr(obj, "nsp_namespace"):
    return  obj.nsp_namespace.lower()

  if type(obj) == list:
    if len(obj) != 2:
      raise Exception("When passing a list to obj_to_namespace it is expected to have two items: The model instance and the field name")
    ns = "%s.%s" % (obj_to_namespace(obj[0]), obj[1].lower())
    return ns

  if hasattr(obj, "db_column") and hasattr(obj, "model"):
    namespace = "%s.%s.%s" % (
      obj.model._meta.app_label, 
      obj.model._meta.model_name,
      obj.name
    )
  elif hasattr(obj, "_meta"):
    namespace = "%s.%s" % (
      obj._meta.app_label, 
      obj._meta.model_name
    )
  if hasattr(obj, "id"):
    namespace = "%s.%s" % (namespace, obj.id)
 

  return namespace.lower()



#############################################################################

def has_perms(user, namespace, level, ambiguous=False):
  
  if type(namespace) not in [str, unicode]:
    namespace = obj_to_namespace(namespace)

  if type(user) != dict:
    if user.is_superuser:
      return True
    perms = load_perms(user)
  else:
    perms = user


  return ((check_perms(perms, namespace, ambiguous=ambiguous) & level) != 0)


#############################################################################

def check_perms(perms, prefix, ambiguous=False):

  """
  Return the user's perms for the specified prefix

  perms <dict> permissions dict
  prefix <string> namespace to check for perms

  ambiguous <bool=False> if True reverse wildcard matching is active and a perm check for a.b.* will
  be matched by the user having perms to a.b.c or a.b.d - only use this if you know what 
  you are doing.
  """

  try:

    token = prefix.split(".")

    i = 1
    l = len(token)
    r = 0

    # collect permission rules with a wildcard in them, so we dont do unecessary
    # regex searches later on
    perms_wc = {}
    for ns, p in perms.items():
      if ns.find("*") > -1:
        perms_wc[re.escape(ns).replace("\*", "[^\.]+")] = p

    while i <= l:
      k = ".".join(token[:i])
      matched = False

      # check for exact match
      if perms.has_key(k):
        r = perms.get(k)

      # check for wildcard matches (if any wildcard rules exist)
      elif perms_wc:
        for ns, p in perms_wc.items():
          a = "^%s$" % ns
          b = "^%s\." % ns
          j = len(a)
          u = len(b)
          if j > matched and re.match(a, k):
            r = p
            matched = j
          elif u > matched and re.match(b, k):
            r = p
            matched = u

      # if not matched at all and ambiguous flag is true, do ambiguous matching
      if not matched and ambiguous:
        m = "^%s" % re.escape(k).replace("\*", "[^\.]+")
        for ns, p in perms.items():
          if re.match(m, ns) and p > r:
            r = p
            break

      i += 1

    return r
  except:
    raise  

###############################################################################

def permissions_apply_additive(data, perms_struct, debug=False, path=[], perms_struct_root=-1, require_explicit=None):
  if not perms_struct:
    return {}

  if type(perms_struct) != dict or type(data) != dict:
    return data

  if perms_struct_root == -1:
    perms_struct_root = perms_struct

  rv = {}
  for k,v in data.items():
    if require_explicit:
      if path:
        path_full = "%s.%s" % (".".join(path), k)
      else:
        path_full = k
      if debug: print "Checking explicit require: %s" % path_full
      r_ex = require_explicit.get(path_full)
      if r_ex:
        #if debug: print "Requiring explicit for %s" % path_full
        if not perms_struct_has_explicit_rule(perms_struct_root, path + [k], r_ex):
          if debug: print "Explicit rule not matched for %s" % path_full
        else:
          if debug: print "Explicit rule matched for %s" % path_full
 
    if perms_struct.get(k):
      rv[k] = permissions_apply_additive(
        v, 
        perms_struct.get(k),
        debug = debug,
        path = path + [k],
        perms_struct_root = perms_struct_root,
        require_explicit = require_explicit
      )
    if perms_struct.get("@%s"%k):
      d = permissions_apply_additive(
        v, 
        perms_struct.get("@%s"%k),
        debug = debug,
        path = path + [k],
        perms_struct_root = perms_struct_root,
        require_explicit = require_explicit
      )
      if type(d) == dict and type(rv.get(k)) == dict:
        rv[k].update(d)
      else:
        rv[k] = d
    if perms_struct.get("*"):
      d = permissions_apply_additive(
        v, 
        perms_struct.get("*"),
        debug = debug,
        path = path + [k],
        perms_struct_root = perms_struct_root,
        require_explicit = require_explicit
      )
      if type(d) == dict and type(rv.get(k)) == dict:
        rv[k].update(d)
      else:
        rv[k] = d

  return rv

#############################################################################

def perms_struct_has_explicit_rule(d, keys, perm):
  a = d
  if type(a) == int:
    return a & perm
  if not keys:
    return False

  r_a = 0 
  r_b = 0
  k = keys[0]
  if a.has_key(k):
    r_a = perms_struct_has_explicit_rule(a[k], keys[1:], perm)
  
  if r_a & perm:
    return True

  if a.has_key("*"):
    r_b = perms_struct_has_explicit_rule(a["*"], keys[1:], perm)

  if r_b & perm:
    return True

  return False
  

def permissions_apply_subtractive(data, perms_struct, debug=False):
  if type(data) != dict:
    raise Exception("Wanted a dict got %s" % type(data))

  for k,p in perms_struct.items():
    if k[0] == "@":
      k = k[1:]

    if data.has_key(k):

      if not p or (type(data[k]) == dict and not any(data[k])):
        del data[k]
      elif type(data[k]) == dict and type(p) == dict:
        permissions_apply_subtractive(
          data.get(k), 
          p, 
          debug=debug
        )
        
    elif k == "*":
      for n,v in data.items():
        if not p or (type(data[n]) == dict and not any(data[n])):
          del data[n]
        elif type(p) == dict and type(v) == dict:
          permissions_apply_subtractive(
            v, 
            p, 
            debug=debug
          )

#############################################################################

def permissions_apply(data, perms_struct, path='', debug=False, ruleset=None):
  if type(perms_struct) != dict:
    load_perms(perms_struct)
    perms_struct = perms_struct._nsp_perms_struct


  ruleset_require = (ruleset or {}).get("require")

  rv = permissions_apply_additive(data, perms_struct, debug=debug, require_explicit=ruleset_require)
  if debug:
    print json.dumps(rv, indent=2)

  permissions_apply_subtractive(rv, perms_struct)
  if debug:
    print json.dumps(rv, indent=2)
    
  return rv

#############################################################################

def permissions_apply_to_serialized_model(smodel, perms_struct):
  inst = smodel.instance
  namespace = obj_to_namespace(inst).split(".")
  structure = d = {}
  l = len(namespace)
  i = 0
  while i < l:
    k = namespace[i]
    if i < l-1:
      d[k] = {}
      d = d[k]
    else:
      d[str(k)] = smodel.data
    i += 1


  r = permissions_apply(
    structure,
    perms_struct
  )


  for k in namespace:
    r = r[k]

  return r

#############################################################################

def perms_structure(perms):
  perms_wc = {}
  for ns, p in perms.items():
    pieces = ns.split(".")
    a = prev = perms_wc 
    n = 0 
    l = len(pieces)
    for k in pieces:
      if n < l-1:
        if not a.has_key(k):
          a[k] = {}
        elif type(a[k]) != dict:
          a["@%s"%k] = a[k]
          a[k] = {}
        a = a[k]
      else:
        if not a.has_key(k):
          a[k] = p
        else:
          a["@%s"%k] = p
      n += 1
   

  return perms_wc 



from setuptools import setup

version = open('config/VERSION').read().strip()
requirements = open('config/requirements.txt').read().split("\n")

setup(
    name='django-namespace-perms',
    version=version,
    author='Twentieth Century',
    author_email='code@20c.com',
    description='granular permission system that allows permissioning for read / write operations all the way down to the field level. Also supports completely arbitrary / custom permission namespaces.',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages=[
      'django_namespace_perms', 
      'django_namespace_perms.auth'
    ],
    url = 'https://github.com/20c/django-namespace-perms',
    download_url = 'https://github.com/20c/django-namespace-perms/%s'%version,
    include_package_data=True,
    install_requires=requirements,
    zip_safe=False
)




from django.contrib import admin
from django.contrib.auth import get_user_model
from django.contrib.auth.admin import UserAdmin as DjangoUserAdmin
from django.contrib.auth.models import User, Group 

from django_namespace_perms.models import GroupPermission, UserPermission
from django_namespace_perms.constants import *
from django_namespace_perms.util import NAMESPACES

from django import forms

import autocomplete_light

def assign_group_to_all_users(modeladmin, request, queryset):
  User = get_user_model()
  users = User.objects.all()
  for group in queryset:
    #note: cant do bulk_save here since right now that doesnt trigger the save signal,
    #revist when/if that get's fixed
    for user in users:
      group.user_set.add(user)
assign_group_to_all_users.short_description = "Assign selected groups to all users"


def revoke_group_from_all_users(modeladmin, request, queryset):
  for group in queryset:
    group.user_set.remove(user)
revoke_group_from_all_users.short_description = "Revoke selected groups from all users"



# Register your models here.

class NamespaceAutocomplete(autocomplete_light.AutocompleteListBase):
  choices = [v for k,v in NAMESPACES]
autocomplete_light.register(NamespaceAutocomplete);

class ManualUserPermissionInline(autocomplete_light.ModelForm):
  class Meta:
    model = UserPermission
    widgets = {
      'namespace' : autocomplete_light.TextWidget('NamespaceAutocomplete', attrs={"style":"width:500px"})
    }
    fields = "__all__"

class ManualGroupPermissionInline(autocomplete_light.ModelForm):
  class Meta:
    model = GroupPermission
    widgets = {
      'namespace' : autocomplete_light.TextWidget('NamespaceAutocomplete', attrs={"style":"width:500px"})
    }
    fields = "__all__"


class GroupPermissionAdmin(admin.ModelAdmin):
  list_display = ('group', 'namespace', 'permissions')

class GroupPermissionInline(admin.TabularInline):
  model = GroupPermission
  form = ManualGroupPermissionInline

class UserPermissionInline(admin.TabularInline):
  model = UserPermission
  form = ManualUserPermissionInline

class GroupAdmin(admin.ModelAdmin):
  list_display = ('name',)
  search_fields = ('name',)
  #remove default permissions form
  exclude = ('permissions',)
  inlines = (GroupPermissionInline,)
  actions = [
    assign_group_to_all_users,
    revoke_group_from_all_users
  ]

class UserAdmin(DjangoUserAdmin):
  inlines = (UserPermissionInline,)
  add_fieldsets = (
    (None, {
        'classes' : ('wide',),
        'fields' : ('username' ,'password1', 'password2', 'email')
    })
  )

  def __init__(self, *args, **kwargs):
    DjangoUserAdmin.__init__(self, *args, **kwargs)
    # remove default permissions forms
    for label, fieldset in self.fieldsets:
      fieldset["fields"] = [x for x in fieldset["fields"] if x not in [
        'user_permissions'
      ]]

try:
  admin.site.unregister(Group)
except:
  pass
admin.site.register(Group, GroupAdmin)
try:
  admin.site.unregister(User)
except:
  pass
admin.site.register(User, UserAdmin)



from django.contrib.auth.models import User;
from django.contrib.auth.backends import ModelBackend

from django_namespace_perms.constants import *
from django_namespace_perms.util import (
  NAMESPACES, 
  obj_to_namespace, 
  permcode_to_namespace, 
  has_perms, 
  load_perms
)

import logging
import re

log = logging.getLogger("django")

# Admin namespace prefixes

# Grant perms to a namespace
ADMIN_NS_GRANT = "admin.grant"

class NSPBackend(ModelBackend):

  """
  Authenticate actions using nsp
  """

  def load_perms(self, user_obj):
    return load_perms(user_obj)

  def has_module_perms(self, user_obj, obj=None):
    if hasattr(obj, "nsp_namespace"):
      fn = getattr(obj, "nsp_namespace")
      if not callable(fn):
        raise Exception("nsp_namespace attribute needs to be callable for %s" % obj)
      namespace = fn()
    else:
      namespace = obj_to_namespace(obj)

    log.info("Checking module perms: %s" % namespace)
    
    perms = self.load_perms(user_obj)

    return has_perms(perms, namespace, PERM_READ)

  def has_perm(self, user_obj, perm, obj=None):

    #if not user_obj.is_authenticated() or not user.is_active:
      #FIXME: load guest perms and proceed
    #  return False

    # super users have access to everything
    if user.is_superuser:
      return True

    namespace, level = permcode_to_namespace(perm)

    write_ops = ["add", "delete", "change"]
    if hasattr(obj, "nsp_write_ops") and callable(obj, "nsp_write_ops"):
      write_ops.extend(getattr(obj, "nsp_write_ops")())

    log.info( "NSP has_perms %s %s %s" % (namespace, perm, level))

    perms = self.load_perms(user_obj)

    return has_perms(perms, namespace, level)
